%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Osm dam
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Autor: Pavel Zupa <pavel.zupa@gmail.com>
%
% Progam hleda nekonfliktni rozestaveni osmi dam pomoci minimalizace konfliktu
%
%
%

% Reprezentace stavu:
% [Y1,Y2,...,Yn] - cisla radku na kterych stoji jednotlive damy
:- use_module(library(lists)).
:- use_module(library(random)).

:- dynamic maxSteps/1.

size(8).

zeroList(0,[]).
zeroList(Length, [0|List]) :-
	L2 is Length-1,
	zeroList(L2,List).

% Zjisti pocet utoku, ktere jsou na kazde policko daneho sloupce
% X - sloupec
% Queens - stav sachovnice
% YList - seznam poctu utoku na policka v danem sloupci
numAttacks(X,Y,Queens,Attacks) :-
	horizAttacks(X,Y,0,Queens,0,Attacks1),
	diag1Attacks(X,Y,0,Queens,Attacks1,Attacks2),
	diag2Attacks(X,Y,0,Queens,Attacks2,Attacks),!.

horizAttacks(_,_,_,[],A,A).
		
horizAttacks(X,Y,XDist,[Q|Queens],Accum,Attacks) :-
	Y is Q,
	X =\= XDist,
	NewAccum is Accum+1,
	XDist2 is XDist+1,
	horizAttacks(X,Y,XDist2,Queens,NewAccum,Attacks);
	XDist2 is XDist+1,
	horizAttacks(X,Y,XDist2,Queens,Accum,Attacks).

diag1Attacks(_,_,_,[],A,A).
diag1Attacks(X,Y,XDist,[Q|Queens],Accum,Attacks) :-
	Y is X-XDist+Q,
	X =\= XDist,
	NewAccum is Accum+1,
	XDist2 is XDist+1,
	diag1Attacks(X,Y,XDist2,Queens,NewAccum,Attacks);
	XDist2 is XDist+1,
	diag1Attacks(X,Y,XDist2,Queens,Accum,Attacks).
diag2Attacks(_,_,_,[],A,A).
diag2Attacks(X,Y,XDist,[Q|Queens],Accum,Attacks) :-
	Q is X-XDist+Y,
	X =\= XDist,
	NewAccum is Accum+1,
	XDist2 is XDist+1,
	diag2Attacks(X,Y,XDist2,Queens,NewAccum,Attacks);
	XDist2 is XDist+1,
	diag2Attacks(X,Y,XDist2,Queens,Accum,Attacks).

printState(Queens) :-
	size(X),
	conflicts(Queens,Conflicts),
	printState(Queens,0,X,Conflicts),
	printLine(Queens),write('+'),nl,!.

printState(_,X,X,_).
printState(Queens,A,Size,Conflicts) :-
	printLine(Queens),write('+'),nl,
	printRow(Queens,A,0,Conflicts),write('|'),nl,
	A2 is A+1,
	printState(Queens,A2,Size,Conflicts).

printRow([],_,_,_).

printRow([A|Queens],A,C,[C|Conflicts]) :-
	write('| X '),
	C1 is C+1,
	printRow(Queens,A,C1,Conflicts).
printRow([A|Queens],A,C,Conflicts) :-
	write('| O '),
	C1 is C+1,
	printRow(Queens,A,C1,Conflicts).
printRow([_|Queens],X,C,[C|Conflicts]) :-
	write('|   '),
	C1 is C+1,
	printRow(Queens,X,C1,Conflicts).
printRow([_|Queens],X,C,Conflicts) :-
	write('|   '),
	C1 is C+1,
	printRow(Queens,X,C1,Conflicts).
printLine([]).
printLine([_|Queens]) :-
	write('+---'),
	printLine(Queens).

% hlavni algoritmus minimalizace konfliktu
minConflict(Queens,_,I,I) :-
	conflicts(Queens,[]).
minConflict(Queens,Max,I,IOut) :-
	I < Max,
	conflicts(Queens, Confs),
	bestMove(Queens,Confs,X,Y),
	nth0(X,Queens,_,Queens2),
	nth0(X,NewQueens,Y,Queens2),
	I1 is I+1,
	write('Step: '),write(I1),nl,
	printState(NewQueens),nl,!,
	minConflict(NewQueens,Max,I1,IOut).

minConflictMute(Queens,_,I,I) :-
	conflicts(Queens,[]).
minConflictMute(Queens,Max,I,IOut) :-
	I < Max,
	conflicts(Queens, Confs),
	bestMove(Queens,Confs,X,Y),
	nth0(X,Queens,_,Queens2),
	nth0(X,NewQueens,Y,Queens2),
	I1 is I+1,!,
	minConflictMute(NewQueens,Max,I1,IOut).

% Conflicts je seznam cisel sloupcu, kde jsou konflikty
conflicts(Queens,Conflicts) :-
	conflicts(Queens,0,Queens,Conflicts),!.
conflicts(_,_,[],[]).
conflicts(Queens,X,[Y|Left],[X|Conflicts]) :-
	numAttacks(X,Y,Queens,Num),
	Num > 0,
	X1 is X+1,
	conflicts(Queens,X1,Left,Conflicts).
conflicts(Queens,X,[Y|Left],Conflicts) :-
	numAttacks(X,Y,Queens,Num),
	Num is 0,
	X1 is X+1,
	conflicts(Queens,X1,Left,Conflicts).

bestMove(Queens,Rows,X,Y) :-
	moves(Queens,Rows,[],Moves),
	size(S),
	bestMove(Moves,m(S,0,0),m(_,X,Y)).

bestMove([],M,M).
bestMove([m(C1,X1,Y1)|Moves],m(C,X,Y),M) :-
	C1 < C,
	bestMove(Moves,m(C1,X1,Y1),M);
	bestMove(Moves,m(C,X,Y),M).

moves(_,[],M,M).
moves(Queens,[R|Rows],OldMoves,Moves) :-
	nth0(R,Queens,Exc),
	incList(Exc,Nums),	
	bagof(m(C,R,Y), (
		member(Y,Nums),
		numAttacks(R,Y,Queens,C)
	),M),
	%write(Nums),write(M),nl,
	append(OldMoves,M,NewMoves),
	moves(Queens,Rows,NewMoves,Moves).

incList(Exclude,List) :-
	size(S),
	incList(0,S,Exclude, List).
incList(S,S,_,[]).
incList(X,S,X,List) :-
	NewX is X+1,
	incList(NewX,S,X,List).
	
incList(X,S,A,[X|List]) :-
	NewX is X+1,
	incList(NewX,S,A,List).

resolve(State,MaxSteps) :-
	write('Input:'),nl,
	printState(State),nl,
	minConflict(State,MaxSteps,0,_),
	write('SUCCESS!!!');
	write('TOO BAD :-(').

resolveMute(State,MaxSteps,I) :-
	minConflictMute(State,MaxSteps,0,I).
randomState(Size,S) :-
	createNums(Size,0,[],S).	
randomState(Size,Conflicts,S) :-
	createNums(Size,0,[],S),
	conflicts(S,C),
	length(C,Conflicts);
	randomState(Size,Conflicts,S),!.
	

createNums(0,_,Nums,Nums) :-!.
createNums(Cnt,Size,Nums,New) :-
	N is Cnt-1,
	Size2 is Size+1,
	random(0,Size2,I),
	nth0(I,Neww,N,Nums),
	createNums(N,Size2,Neww,New).

test(Count,Conflicts) :-
	write('Initial conflicts: '),write(Conflicts),nl,
	write('Steps to solve: '),
	doTest(Count,Conflicts).
	
doTest(1,Conflicts):-
	randomState(8,Conflicts,State),
	resolveMute(State,40,Steps),
	write(Steps),nl;
	write('X').
	
doTest(Count,Conflicts) :-
	randomState(8,Conflicts,State),
	resolveMute(State,40,Steps),
	write(Steps),write(', '),
	Count2 is Count-1,
	doTest(Count2,Conflicts);
	write('X, '),
	Count2 is Count-1,
	doTest(Count2,Conflicts).
	

